@@ -18,11 +18,31 @@
  */
 package org.apache.felix.fileinstall.internal;
 
-import java.io.*;
+import java.io.BufferedInputStream;
+import java.io.File;
+import java.io.FileInputStream;
+import java.io.FileOutputStream;
+import java.io.IOException;
+import java.io.InputStream;
+import java.io.InputStreamReader;
+import java.io.OutputStreamWriter;
+import java.io.Reader;
+import java.io.Writer;
+import java.lang.reflect.Array;
+import java.lang.reflect.InvocationTargetException;
+import java.lang.reflect.Method;
 import java.net.URI;
 import java.security.AccessController;
 import java.security.PrivilegedAction;
-import java.util.*;
+import java.util.ArrayList;
+import java.util.Dictionary;
+import java.util.Enumeration;
+import java.util.HashMap;
+import java.util.Hashtable;
+import java.util.List;
+import java.util.Map;
+import java.util.Properties;
+import java.util.Set;
 
 import org.apache.felix.fileinstall.ArtifactInstaller;
 import org.apache.felix.fileinstall.ArtifactListener;
@@ -30,8 +50,13 @@ import org.apache.felix.fileinstall.internal.Util.Logger;
 import org.apache.felix.utils.collections.DictionaryAsMap;
 import org.apache.felix.utils.properties.InterpolationHelper;
 import org.apache.felix.utils.properties.TypedProperties;
-import org.osgi.framework.*;
-import org.osgi.service.cm.*;
+import org.osgi.framework.BundleContext;
+import org.osgi.framework.Constants;
+import org.osgi.framework.ServiceRegistration;
+import org.osgi.service.cm.Configuration;
+import org.osgi.service.cm.ConfigurationAdmin;
+import org.osgi.service.cm.ConfigurationEvent;
+import org.osgi.service.cm.ConfigurationListener;
 
 /**
  * ArtifactInstaller for configurations.
@@ -43,13 +68,82 @@ public class ConfigInstaller implements ArtifactInstaller, ConfigurationListener
     private final ConfigurationAdmin configAdmin;
     private final FileInstall fileInstall;
     private final Map<String, String> pidToFile = new HashMap<>();
-    private ServiceRegistration registration;
+    private final Method getFactoryConfigurationMethod;
+    private final Method addAttributesMethod;
+    private final Method getAttributesMethod;
+    private final Method removeAttributesMethod;
+    private final Method updateIfDifferentMethod;
+    private final Object READ_ONLY_ATTRIBUTE_ARRAY;
+    private ServiceRegistration<?> registration;
 
+    @SuppressWarnings("unchecked")
     ConfigInstaller(BundleContext context, ConfigurationAdmin configAdmin, FileInstall fileInstall)
     {
         this.context = context;
         this.configAdmin = configAdmin;
         this.fileInstall = fileInstall;
+
+        Method gfcMethod = null;
+        Method aaMethod = null;
+        Method gaMethod = null;
+        Method raMethod = null;
+        Method uidMethod = null;
+
+        if (this.configAdmin != null) {
+            for (Method method : ConfigurationAdmin.class.getDeclaredMethods())
+            {
+                if ("getFactoryConfiguration".equals(method.getName()) && (method.getParameterCount() == 3))
+                {
+                    gfcMethod = method;
+                }
+            }
+            for (Method method : Configuration.class.getDeclaredMethods())
+            {
+                if ("addAttributes".equals(method.getName()))
+                {
+                    aaMethod = method;
+                }
+                else if ("getAttributes".equals(method.getName()))
+                {
+                    gaMethod = method;
+                }
+                else if ("removeAttributes".equals(method.getName()))
+                {
+                    raMethod = method;
+                }
+                else if ("updateIfDifferent".equals(method.getName()))
+                {
+                    uidMethod = method;
+                }
+            }
+        }
+
+        this.getFactoryConfigurationMethod = gfcMethod;
+        this.addAttributesMethod = aaMethod;
+        this.getAttributesMethod = gaMethod;
+        this.removeAttributesMethod = raMethod;
+        this.updateIfDifferentMethod = uidMethod;
+
+        Object roAttributesArray = null;
+        if (this.addAttributesMethod != null)
+        {
+            try {
+                @SuppressWarnings("rawtypes")
+                Class<? extends Enum> attr = (Class<? extends Enum>)context.getBundle()
+                    .loadClass("org.osgi.service.cm.Configuration$ConfigurationAttribute");
+
+                Object attrArray = Array.newInstance(attr, 1);
+                Array.set(attrArray, 0, Enum.valueOf(attr, "READ_ONLY"));
+
+                roAttributesArray = attrArray;
+            }
+            catch (ClassNotFoundException cnfe) {
+                throw new IllegalStateException(
+                    "Cannot find the enum org.osgi.service.cm.Configuration.ConfigurationAttribute", cnfe);
+            }
+        }
+
+        this.READ_ONLY_ATTRIBUTE_ARRAY = roAttributesArray;
     }
 
     public void init()
@@ -63,12 +157,12 @@ public class ConfigInstaller implements ArtifactInstaller, ConfigurationListener
                 this, null);
         try
         {
-            Configuration[] configs = configAdmin.listConfigurations(null);
+            Configuration[] configs = getConfigurationAdmin().listConfigurations(null);
             if (configs != null)
             {
                 for (Configuration config : configs)
                 {
-                    Dictionary dict = config.getProperties();
+                    Dictionary<?,?> dict = config.getProperties();
                     String fileName = dict != null ? (String) dict.get( DirectoryWatcher.FILENAME ) : null;
                     if (fileName != null)
                     {
@@ -147,11 +241,17 @@ public class ConfigInstaller implements ArtifactInstaller, ConfigurationListener
                 Configuration config = getConfigurationAdmin().getConfiguration(
                                             configurationEvent.getPid(),
                                             "?");
-                Dictionary dict = config.getProperties();
+                Dictionary<?,?> dict = config.getProperties();
                 String fileName = dict != null ? (String) dict.get( DirectoryWatcher.FILENAME ) : null;
                 File file = fileName != null ? fromConfigKey(fileName) : null;
                 if( file != null && file.isFile() && canHandle( file ) ) {
                     pidToFile.put(config.getPid(), fileName);
+                    if (!file.canWrite())
+                    {
+                        Util.log(context, Logger.LOG_WARNING, "File is not writeable: " + fileName, null);
+                        return;
+                    }
+
                     TypedProperties props = new TypedProperties( bundleSubstitution() );
                     try (Reader r = new InputStreamReader(new FileInputStream(file), encoding()))
                     {
@@ -168,7 +268,7 @@ public class ConfigInstaller implements ArtifactInstaller, ConfigurationListener
                             propertiesToRemove.add(key);
                         }
                     }
-                    for( Enumeration e  = dict.keys(); e.hasMoreElements(); )
+                    for( Enumeration<?> e  = dict.keys(); e.hasMoreElements(); )
                     {
                         String key = e.nextElement().toString();
                         if( !Constants.SERVICE_PID.equals(key)
@@ -194,7 +294,7 @@ public class ConfigInstaller implements ArtifactInstaller, ConfigurationListener
             }
             catch (Exception e)
             {
-                Util.log( context, Logger.LOG_INFO, "Unable to save configuration", e );
+                Util.log( context, Logger.LOG_ERROR, "Unable to save configuration", e );
             }
         }
 
@@ -286,30 +386,39 @@ public class ConfigInstaller implements ArtifactInstaller, ConfigurationListener
         String pid[] = parsePid(f.getName());
         Configuration config = getConfiguration(toConfigKey(f), pid[0], pid[1]);
 
+        clearReadOnlyIfWritable(pid, config, f);
+
         Dictionary<String, Object> props = config.getProperties();
         Hashtable<String, Object> old = props != null ? new Hashtable<String, Object>(new DictionaryAsMap<>(props)) : null;
         if (old != null) {
-        	old.remove( DirectoryWatcher.FILENAME );
-        	old.remove( Constants.SERVICE_PID );
-        	old.remove( ConfigurationAdmin.SERVICE_FACTORYPID );
+            old.remove( DirectoryWatcher.FILENAME );
+            old.remove( Constants.SERVICE_PID );
+            old.remove( ConfigurationAdmin.SERVICE_FACTORYPID );
         }
 
-        if( !ht.equals( old ) )
-        {
-            ht.put(DirectoryWatcher.FILENAME, toConfigKey(f));
-            if (old == null) {
-                Util.log(context, Logger.LOG_INFO, "Creating configuration from " + pid[0]
-                        + (pid[1] == null ? "" : "-" + pid[1]) + ".cfg", null);
-            } else {
-                Util.log(context, Logger.LOG_INFO, "Updating configuration from " + pid[0]
-                        + (pid[1] == null ? "" : "-" + pid[1]) + ".cfg", null);
+        try {
+            if( !ht.equals( old ) )
+            {
+                ht.put(DirectoryWatcher.FILENAME, toConfigKey(f));
+                if (old == null) {
+                    Util.log(context, Logger.LOG_INFO, "Creating configuration {"
+                            + config.getPid()
+                            + "} from " + f.getAbsolutePath(), null);
+                } else {
+                    Util.log(context, Logger.LOG_INFO, "Updating configuration {"
+                            + config.getPid()
+                            + "} from " + f.getAbsolutePath(), null);
+                }
+                update0(pid, config, ht, f);
+                return true;
+            }
+            else
+            {
+                return false;
             }
-            config.update(ht);
-            return true;
         }
-        else
-        {
-            return false;
+        finally {
+            setReadOnlyInNotWritable(pid, config, f);
         }
     }
 
@@ -324,9 +433,10 @@ public class ConfigInstaller implements ArtifactInstaller, ConfigurationListener
     boolean deleteConfig(File f) throws Exception
     {
         String pid[] = parsePid(f.getName());
-        Util.log(context, Logger.LOG_INFO, "Deleting configuration from " + pid[0]
-                + (pid[1] == null ? "" : "-" + pid[1]) + ".cfg", null);
         Configuration config = getConfiguration(toConfigKey(f), pid[0], pid[1]);
+        Util.log(context, Logger.LOG_INFO, "Deleting configuration {"
+                + config.getPid()
+                + "} from " + f.getAbsolutePath(), null);
         config.delete();
         return true;
     }
@@ -366,7 +476,7 @@ public class ConfigInstaller implements ArtifactInstaller, ConfigurationListener
     {
         Configuration oldConfiguration = findExistingConfiguration(fileName);
         Configuration cachedConfiguration = oldConfiguration != null ?
-                getConfigurationAdmin().getConfiguration(oldConfiguration.getPid()) : null;
+                getConfigurationAdmin().getConfiguration(oldConfiguration.getPid(), null) : null;
         if (cachedConfiguration != null)
         {
             return cachedConfiguration;
@@ -376,7 +486,12 @@ public class ConfigInstaller implements ArtifactInstaller, ConfigurationListener
             Configuration newConfiguration;
             if (factoryPid != null)
             {
-                newConfiguration = getConfigurationAdmin().createFactoryConfiguration(pid, "?");
+                if (getFactoryConfigurationMethod != null) {
+                    newConfiguration = (Configuration)getFactoryConfigurationMethod.invoke(getConfigurationAdmin(), pid, factoryPid, "?");
+                }
+                else {
+                    newConfiguration = getConfigurationAdmin().createFactoryConfiguration(pid, "?");
+                }
             }
             else
             {
@@ -410,6 +525,139 @@ public class ConfigInstaller implements ArtifactInstaller, ConfigurationListener
         };
     }
 
+    @SuppressWarnings({ "rawtypes", "unchecked" })
+    boolean isReadOnly(Configuration configuration) throws IOException {
+        if (getAttributesMethod == null) {
+            return false;
+        }
+
+        try {
+            Set<? extends Enum> attributes = (Set<? extends Enum>)getAttributesMethod.invoke(configuration);
+
+            if (attributes.isEmpty()) {
+                return false;
+            }
+
+            return attributes.iterator().next().name().equals("READ_ONLY");
+        }
+        catch (InvocationTargetException ite) {
+            Throwable targetException = ite.getTargetException();
+            if (targetException instanceof IOException) {
+                throw (IOException)targetException;
+            }
+
+            throw new RuntimeException(targetException);
+        }
+        catch (Throwable t) {
+            throw new RuntimeException(t);
+        }
+    }
+
+    void clearReadOnlyIfWritable(String[] pid, Configuration configuration, File f) throws IOException {
+        if (removeAttributesMethod == null) {
+            return;
+        }
+
+        try {
+            if (Util.canWrite(f) && isReadOnly(configuration)) {
+                Util.log(context, Logger.LOG_INFO, "Removing  READ_ONLY attribute from configuration {"
+                        + configuration.getPid()
+                        + "} from " + f.getAbsolutePath(), null);
+                removeAttributesMethod.invoke(configuration, READ_ONLY_ATTRIBUTE_ARRAY);
+            }
+        }
+        catch (InvocationTargetException ite) {
+            Throwable targetException = ite.getTargetException();
+            if (targetException instanceof IOException) {
+                throw (IOException)targetException;
+            }
+
+            throw new RuntimeException(targetException);
+        }
+        catch (Throwable t) {
+            throw new RuntimeException(t);
+        }
+    }
+
+    void clearReadOnlyToUpdate(String[] pid, Configuration configuration, File f) throws IOException {
+        if (removeAttributesMethod == null) {
+            return;
+        }
+
+        try {
+            if (isReadOnly(configuration)) {
+                Util.log(context, Logger.LOG_INFO, "Temporarily removing  READ_ONLY attribute from configuration for external update {"
+                        + configuration.getPid()
+                        + "} from " + f.getAbsolutePath(), null);
+                removeAttributesMethod.invoke(configuration, READ_ONLY_ATTRIBUTE_ARRAY);
+            }
+        }
+        catch (InvocationTargetException ite) {
+            Throwable targetException = ite.getTargetException();
+            if (targetException instanceof IOException) {
+                throw (IOException)targetException;
+            }
+
+            throw new RuntimeException(targetException);
+        }
+        catch (Throwable t) {
+            throw new RuntimeException(t);
+        }
+    }
+
+    boolean setReadOnlyInNotWritable(String[] pid, Configuration configuration, File f) throws IOException {
+        if (addAttributesMethod == null) {
+            return false;
+        }
+
+        try {
+            if (!Util.canWrite(f)) {
+                Util.log(context, Logger.LOG_INFO, "Adding  READ_ONLY attribute to configuration {"
+                        + configuration.getPid()
+                        + "} from " + f.getAbsolutePath(), null);
+                addAttributesMethod.invoke(configuration, READ_ONLY_ATTRIBUTE_ARRAY);
+
+                return true;
+            }
+        }
+        catch (InvocationTargetException ite) {
+            Throwable targetException = ite.getTargetException();
+            if (targetException instanceof IOException) {
+                throw (IOException)targetException;
+            }
+
+            throw new RuntimeException(targetException);
+        }
+        catch (Throwable t) {
+            throw new RuntimeException(t);
+        }
+
+        return false;
+    }
+
+    void update0(String[] pid, final Configuration config, final Hashtable<String, Object> ht, File f) throws IOException {
+        clearReadOnlyToUpdate(pid, config, f);
+        if (updateIfDifferentMethod != null) {
+            try {
+                updateIfDifferentMethod.invoke(config, ht);
+            }
+            catch (InvocationTargetException ite) {
+                Throwable targetException = ite.getTargetException();
+                if (targetException instanceof IOException) {
+                    throw (IOException)targetException;
+                }
+
+                throw new RuntimeException(targetException);
+            }
+            catch (Throwable t) {
+                throw new RuntimeException(t);
+            }
+        }
+        else {
+            config.update(ht);
+        }
+    }
+
     private String escapeFilterValue(String s) {
         return s.replaceAll("[(]", "\\\\(").
                 replaceAll("[)]", "\\\\)").
